/*
 * Copyright 1993-2009 NVIDIA Corporation.  All rights reserved.
 *
 * NVIDIA Corporation and its licensors retain all intellectual property and 
 * proprietary rights in and to this software and related documentation and 
 * any modifications thereto.  Any use, reproduction, disclosure, or distribution 
 * of this software and related documentation without an express license 
 * agreement from NVIDIA Corporation is strictly prohibited.
 * 
 */
#pragma warning(disable: 4312)

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#ifdef _WIN32
#define WINDOWS_LEAN_AND_MEAN
#include <windows.h>
#endif
#include <d3dx9.h>

#include <cuda.h>
#include <builtin_types.h>
#include <cuda_runtime_api.h>

#include <cufft.h>
#include <cuda_d3d9_interop.h>
#include <cutil_inline.h>

#include "fluidsD3D9_kernels.h"
#include <rendercheck_d3d9.h>

#define MAX_EPSILON 10

// CUDA example code that implements the frequency space version of 
// Jos Stam's paper 'Stable Fluids' in 2D. This application uses the 
// CUDA FFT library (CUFFT) to perform velocity diffusion and to 
// force non-divergence in the velocity field at each time step. It uses 
// CUDA-OpenGL interoperability to update the particle field directly
// instead of doing a copy to system memory before drawing. Texture is
// used for automatic bilinear interpolation at the velocity advection step. 

HWND                    hWnd;              // Window handle 
LPDIRECT3D9             g_pD3D       = NULL; // Used to create the D3DDevice
unsigned int            g_iAdapter   = NULL; // Adapter
LPDIRECT3DDEVICE9       g_pd3dDevice = NULL; // Rendering device
LPDIRECT3DVERTEXBUFFER9 g_pVB        = NULL; // Buffer to hold particles
LPDIRECT3DTEXTURE9      g_pTexture   = NULL; // Texture to render points

#define D3DFVF_CUSTOMVERTEX (D3DFVF_XYZ|D3DFVF_DIFFUSE)
void updateVB(void);
void initParticles(cData *p, int dx, int dy);

// CUFFT plan handle
static cufftHandle g_planr2c;
static cufftHandle g_planc2r;
static cData *g_vxfield = NULL;
static cData *g_vyfield = NULL;

cData *g_hvfield = NULL;
cData *g_dvfield = NULL;
static int wWidth = max(512,DIM);
static int wHeight = max(512,DIM);

static int clicked = 0;
static int fpsCount = 0;
static int fpsLimit = 1;
unsigned int timer;

// Particle data
static Vertex *g_mparticles = NULL;  
static cData  *g_particles    = NULL;
static int lastx = 0, lasty = 0;

// Texture pitch
//unsigned int g_tPitch = 0;
size_t g_tPitch = 0;

bool				  g_bQAReadback = false;
int					  g_iFrameToCompare = 100;
bool				  g_bQAAddTestForce = true;

VOID Cleanup() 
{ 

    // Unregister vertex buffer
	cudaD3D9UnregisterResource(g_pVB);
    cutilCheckMsg("cudaD3D9UnregisterResource failed");

    unbindTexture();
    deleteTexture();
    
    // Free all host and device resources
    free(g_hvfield); free(g_particles); 
    cudaFree(g_dvfield); 
    cudaFree(g_vxfield); cudaFree(g_vyfield);

	cufftDestroy(g_planr2c);
    cufftDestroy(g_planc2r);

    if (g_pVB != NULL) {        
        g_pVB->Release();
        g_pVB = NULL;
    }
    if (g_pTexture != NULL) {        
        g_pTexture->Release();
        g_pTexture = NULL;
    }

    // Uninitialize CUDA
    cudaThreadExit();
    cutilCheckMsg("cudaThreadExit failed");

    if (g_pd3dDevice != NULL) {
        g_pd3dDevice->Release();
        g_pd3dDevice = NULL;
    }
    if (g_pD3D != NULL) {       
        g_pD3D->Release();  
        g_pD3D = NULL;
    }

    cutilCheckError(cutDeleteTimer(timer));  

}

LRESULT WINAPI MsgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) {
    switch (msg) {
        case WM_DESTROY:
            Cleanup();
            PostQuitMessage(0);
            return 0;
        case WM_KEYDOWN:
            switch(wParam) {
                case 27:
                Cleanup();
                PostQuitMessage(0);
                break;
                case 0x52:
                    memset(g_hvfield, 0, sizeof(cData) * DS);
                    cudaMemcpy(g_dvfield, g_hvfield, sizeof(cData) * DS, 
                               cudaMemcpyHostToDevice);

                    initParticles(g_particles, DIM, DIM);
                    updateVB();
                    break;
                default:
                break;
            }
            break;
 		case WM_SIZE:
			wWidth = LOWORD(lParam);
			wHeight = HIWORD(lParam);
			break;
        case WM_MOUSEMOVE: 
            if (wParam == MK_LBUTTON) clicked = 1;
            else clicked = 0;
            int x = LOWORD(lParam), y = HIWORD(lParam);
             
            // Convert motion coordinates to domain
            float fx = (x / (float)wWidth);        
            float fy = (y / (float)wHeight);
            int nx = (int)(fx * DIM);        
            int ny = (int)(fy * DIM);   
           
            if (clicked && nx < DIM-FR && nx > FR-1 && ny < DIM-FR && ny > FR-1) 
			{
                int ddx = LOWORD(lParam) - lastx;
                int ddy = HIWORD(lParam) - lasty;
                
                fx = ddx / (float)wWidth;
                fy = ddy / (float)wHeight;
                int spy = ny-FR;
                int spx = nx-FR;
                addForces(g_dvfield, DIM, DIM, spx,  spy, FORCE * DT * fx, FORCE * DT * fy, FR, g_tPitch);
                lastx = x; lasty = y;
            } 
            break;
    }
    return DefWindowProc( hWnd, msg, wParam, lParam );
 }

HRESULT InitVB() {
    // Create the vertex buffer.
    if (FAILED(g_pd3dDevice->CreateVertexBuffer(DS * sizeof(Vertex),
                                              0, D3DFVF_CUSTOMVERTEX,
                                              D3DPOOL_DEFAULT, &g_pVB, NULL)))
        return E_FAIL;


	// Initialize the Vertex Buffer with the particles
    updateVB();

	cudaD3D9RegisterResource(g_pVB, cudaD3D9RegisterFlagsNone);
	cutilCheckMsg("cudaD3D9RegisterResource failed");

    return S_OK;
}

HRESULT InitPointTexture() {
	// Create the texture.
    int width = 64;
    int height = width;
    if (FAILED(D3DXCreateTexture(g_pd3dDevice,
                                 width, height, D3DX_DEFAULT,
                                 D3DUSAGE_AUTOGENMIPMAP, D3DFMT_A8R8G8B8, D3DPOOL_MANAGED,
                                 &g_pTexture)))
        return E_FAIL;

    // Fill in top level
    D3DLOCKED_RECT rect;
    if (FAILED(g_pTexture->LockRect(0, &rect, 0, 0)))
        return E_FAIL;
    typedef unsigned int TexelType;
    TexelType* texel = (TexelType*)rect.pBits;
    for (int y = - height / 2; y < height / 2; ++y) {
        float yf = y + 0.5f;
        TexelType* t = texel;
        for (int x = - width / 2; x < width / 2; ++x) {
            float xf = x + 0.5f;
            float radius = (float)width / 32;
            float dist = sqrtf(xf * xf + yf * yf) / radius;
            float n = 0.1f;
            float value;
            if (dist < 1)
                value = 1 - 0.5f * powf(dist, n);
            else if (dist < 2)
                value = 0.5f * powf(2 - dist, n);
            else
                value = 0;
            value *= 75;
            unsigned char* c = (unsigned char*)t;
            c[0] = c[1] = c[2] = c[3] = (unsigned char)value;
            ++t;
        }
        texel += rect.Pitch / sizeof(TexelType);
    }
    if (FAILED(g_pTexture->UnlockRect(0)))
        return E_FAIL;

    // Set sampler state
    if (FAILED(g_pd3dDevice->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR)))
        return E_FAIL;
    if (FAILED(g_pd3dDevice->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR)))
        return E_FAIL;
    
    return S_OK;
}

void updateVB(void) {
    Vertex *data = new Vertex[DS];
	g_pVB->Lock(0,DS * sizeof(Vertex),(void**)&data,0);
    for (int i = 0; i < DS; i++) {
        data[i].x = g_particles[i].x;
        data[i].y = g_particles[i].y;
        data[i].z = 0.f;
        data[i].c = 0xff00ff00;
    }
	g_pVB->Unlock();
}

HRESULT InitD3D(HWND hWnd) 
{
    // Create the D3D object.
    if(NULL == (g_pD3D = Direct3DCreate9(D3D_SDK_VERSION)))
        return E_FAIL;
    
    // Find the first CUDA capable device
    for(g_iAdapter = 0; g_iAdapter < g_pD3D->GetAdapterCount(); g_iAdapter++)
    {
		D3DCAPS9 caps;
		if (FAILED(g_pD3D->GetDeviceCaps(g_iAdapter, D3DDEVTYPE_HAL, &caps)))
			// Adapter doesn't support Direct3D
			continue;
        D3DADAPTER_IDENTIFIER9 ident;
        int device;
        g_pD3D->GetAdapterIdentifier(g_iAdapter, 0, &ident);
        cudaD3D9GetDevice(&device, ident.DeviceName);
        if (cudaSuccess == cudaGetLastError() )
            break;
    }

    // we check to make sure we have found a cuda-compatible D3D device to work on
    if(g_iAdapter == g_pD3D->GetAdapterCount() ) {
        printf("No CUDA-compatible Direct3D device available\nPASSED\n");
        return E_FAIL;
    }

    // Set up the structure used to create the D3DDevice
    D3DPRESENT_PARAMETERS d3dpp;
    ZeroMemory( &d3dpp, sizeof(d3dpp) );
    d3dpp.Windowed = TRUE;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;
    
    // Create the D3DDevice
    if ( FAILED( g_pD3D->CreateDevice( g_iAdapter, D3DDEVTYPE_HAL, hWnd,
                                     D3DCREATE_HARDWARE_VERTEXPROCESSING,
                                     &d3dpp, &g_pd3dDevice ) ) )
        return E_FAIL;    

	// Direct3D and CUDA can only interoperate on the same device.  We me pass the D3D9 device
	// to CUDA via the new CUDA 2.0 interface (required for XP and Vista to work)
	cudaError_t error;
	error = cudaD3D9SetDirect3DDevice(g_pd3dDevice);
    cutilCheckMsg("cudaD3D9SetDirect3DDevice failed");

    // Set projection matrix
    D3DXMATRIXA16 matProj;
    D3DXMatrixOrthoOffCenterLH( &matProj, 0, 1, 1, 0, 0, 1);
    g_pd3dDevice->SetTransform( D3DTS_PROJECTION, &matProj );
   
    // Turn off D3D lighting, since we are providing our own vertex colors
    if( FAILED( g_pd3dDevice->SetRenderState( D3DRS_LIGHTING, FALSE ) ) )
        return E_FAIL;
        
    return S_OK;
}

HRESULT InitCUFFT()
{
	// You can only call CUDA D3D9 device has been bound to the CUDA
	// context, otherwise it will not work
	g_hvfield = (cData*)malloc(sizeof(cData) * DS);
	memset(g_hvfield, 0, sizeof(cData) * DS);
  
	// Allocate and initialize device data
	cudaMallocPitch((void**)&g_dvfield, &g_tPitch, sizeof(cData)*DIM, DIM);
    
	cudaMemcpy(g_dvfield, g_hvfield, sizeof(cData) * DS, 
			   cudaMemcpyHostToDevice); 

	// Temporary complex velocity field data     
	cudaMalloc((void**)&g_vxfield, sizeof(cData) * PDS);
	cudaMalloc((void**)&g_vyfield, sizeof(cData) * PDS);
    
	setupTexture(DIM, DIM);
	bindTexture();
    
	// Create particle array
	g_particles = (cData*)malloc(sizeof(cData) * DS);
	memset(g_particles, 0, sizeof(cData) * DS);   
    
	initParticles(g_particles, DIM, DIM); 

	// Create CUFFT transform plan configuration
	cufftPlan2d(&g_planr2c, DIM, DIM, CUFFT_R2C);
	cufftPlan2d(&g_planc2r, DIM, DIM, CUFFT_C2R);

	return S_OK;
}

void display(void) 
{  
	cutilCheckError(cutStartTimer(timer));  

	advectVelocity(g_dvfield, (float*)g_vxfield, (float*)g_vyfield, DIM, RPADW, DIM, DT, g_tPitch);
	{
	    // Forward FFT
		cufftExecR2C(g_planr2c, (cufftReal*)g_vxfield, (cufftComplex*)g_vxfield); 
		cufftExecR2C(g_planr2c, (cufftReal*)g_vyfield, (cufftComplex*)g_vyfield);
	
		diffuseProject(g_vxfield, g_vyfield, CPADW, DIM, DT, VIS, g_tPitch);

		// Inverse FFT
		cufftExecC2R(g_planc2r, (cufftComplex*)g_vxfield, (cufftReal*)g_vxfield); 
		cufftExecC2R(g_planc2r, (cufftComplex*)g_vyfield, (cufftReal*)g_vyfield);
	}
	updateVelocity(g_dvfield, (float*)g_vxfield, (float*)g_vyfield, DIM, RPADW, DIM, g_tPitch);

	// Map D3D9 vertex buffer to CUDA
	{
		cudaD3D9MapResources(1, (IDirect3DResource9 **)&g_pVB);
		cutilCheckMsg("cudaD3D9MapResource failed");
		// This gets a pointer from the Vertex Buffer
		cudaD3D9ResourceGetMappedPointer( (void **)&g_mparticles, g_pVB, 0, 0 );
		cutilCheckMsg("cudaD3D9ResourceGetMappedPointer failed");

		advectParticles(g_mparticles, g_dvfield, DIM, DIM, DT, g_tPitch);

		// Unmap vertex buffer
		cudaD3D9UnmapResources(1, (IDirect3DResource9 **)&g_pVB);
		cutilCheckMsg("cudaD3D9UnmapResources failed");
	}

   g_pd3dDevice->Clear(0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB(0,0,0), 1.0f, 0);
   g_pd3dDevice->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);
   g_pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
   g_pd3dDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ONE);
   g_pd3dDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);
   g_pd3dDevice->SetRenderState(D3DRS_POINTSPRITEENABLE, TRUE);
   float size = 16;
   g_pd3dDevice->SetRenderState(D3DRS_POINTSIZE, *((DWORD*)&size));
   g_pd3dDevice->SetTexture(0, g_pTexture);

   if(SUCCEEDED(g_pd3dDevice->BeginScene())) 
   {
   
       // Draw particles
       g_pd3dDevice->SetStreamSource(0, g_pVB, 0, sizeof(Vertex));
       g_pd3dDevice->SetFVF(D3DFVF_CUSTOMVERTEX);
       g_pd3dDevice->DrawPrimitive(D3DPT_POINTLIST, 0, DS);
    
       g_pd3dDevice->EndScene();
   }

   // Finish timing before swap buffers to avoid refresh sync
   cutilCheckError(cutStopTimer(timer));  
   // Present the backbuffer contents to the display
   g_pd3dDevice->Present(NULL, NULL, NULL, NULL);

   fpsCount++;
   if (fpsCount == fpsLimit) {
       char fps[256];
       float ifps = 1.f / (cutGetAverageTimerValue(timer) / 1000.f);
       sprintf(fps, "Cuda/D3D Stable Fluids (%d x %d): %3.1f fps", DIM, DIM, ifps);  
       SetWindowText(hWnd, fps); 
       fpsCount = 0; 
       fpsLimit = (int)max(ifps, 1.f);
       cutilCheckError(cutResetTimer(timer));  
    }
}

void initParticles(cData *p, int dx, int dy) {
    int i, j;
    for (i = 0; i < dy; i++) {
        for (j = 0; j < dx; j++) {
            p[i*dx+j].x = ((j+0.5f)/dx) + 
                          (rand() / (float)RAND_MAX - 0.5f) / dx;
            p[i*dx+j].y = ((i+0.5f)/dy) + 
                          (rand() / (float)RAND_MAX - 0.5f) / dy;
                          
        }
    }
}

int main(int argc, char** argv) 
{
    cutilCheckError(cutCreateTimer(&timer));
    cutilCheckError(cutResetTimer(timer));  
    
	// command line options
    if (argc > 1) 
	{
		// automatied build testing harness
        if (cutCheckCmdLineFlag(argc, (const char **)argv, "qatest") ||
			cutCheckCmdLineFlag(argc, (const char **)argv, "noprompt"))
		{
            g_bQAReadback = true;
        }
	}

    HINSTANCE hInst = GetModuleHandle(NULL);
    // Register the window class
    WNDCLASSEX wc = { sizeof(WNDCLASSEX), CS_CLASSDC, MsgProc, 0L, 0L,
                      GetModuleHandle(NULL), NULL, NULL, NULL, NULL,
                      "fluidsD3D", NULL };
    RegisterClassEx(&wc);

    // Create the application's window
	int xBorder = ::GetSystemMetrics(SM_CXSIZEFRAME);
	int yMenu = ::GetSystemMetrics(SM_CYMENU);
	int yBorder = ::GetSystemMetrics(SM_CYSIZEFRAME);
    hWnd = CreateWindow("fluidsD3D", "Cuda/D3D Stable Fluids",
                        WS_OVERLAPPEDWINDOW, 100, 100, wWidth + 2*xBorder, wHeight+ 2*yBorder+yMenu,
                        NULL, NULL, wc.hInstance, NULL );

    if (SUCCEEDED(InitD3D(hWnd)) && 
		SUCCEEDED(InitCUFFT()) &&
		SUCCEEDED(InitVB()) && 
		SUCCEEDED(InitPointTexture())) 
	{
		// Initialize interoperability between CUDA and Direct3D
		// Register vertex buffer with CUDA
		cudaD3D9ResourceSetMapFlags(g_pVB, 0);
		cutilCheckMsg("cudaD3D9ResourceSetMapFlags failed");

		ShowWindow(hWnd, SW_SHOWDEFAULT);
        UpdateWindow(hWnd);

        // Rendering loop
        MSG msg;
        ZeroMemory(&msg, sizeof(msg));
        while(msg.message!=WM_QUIT) 
		{
            if(PeekMessage(&msg, NULL, 0U, 0U, PM_REMOVE)) 
			{
                TranslateMessage(&msg);
                DispatchMessage(&msg);
            } 
			else
			{ 
				display();    

				if(g_bQAReadback)
				{
					for(int count=0;count<g_iFrameToCompare;count++)
					{
						// add in a little force so the automated testing is interesing.
						if(g_bQAReadback && g_bQAAddTestForce) 
						{
							int x = wWidth/(count+1); int y = wHeight/(count+1);
							float fx = (x / (float)wWidth);        
							float fy = (y / (float)wHeight);
							int nx = (int)(fx * DIM);        
							int ny = (int)(fy * DIM);   

							int ddx = 35;
							int ddy = 35;
							fx = ddx / (float)wWidth;
							fy = ddy / (float)wHeight;
							int spy = ny-FR;
							int spx = nx-FR;

							addForces(g_dvfield, DIM, DIM, spx,  spy, FORCE * DT * fx, FORCE * DT * fy, FR, g_tPitch);
							//g_bQAAddTestForce = false; // only add it once
						}

						display();    					
					}

					const char *ref_image_path = "ref_fluidsD3D9.ppm";
					const char *cur_image_path = "qatest_fluidsD3D9.ppm";

				    // Save a reference of our current test run image
				    CheckRenderD3D9::BackbufferToPPM(g_pd3dDevice,cur_image_path);

				    // compare to offical reference image, printing PASS or FAIL.
				    bool bPassed = CheckRenderD3D9::PPMvsPPM(cur_image_path,ref_image_path,argv[0],MAX_EPSILON,0.30f);

					PostQuitMessage(0);
				}

			}
        }
    } 
    UnregisterClass("fluidsD3D", wc.hInstance );
}
